@@ -1,18 +1,21 @@
 /* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2; -*- */
-/* *  Copyright © 2011 Igalia S.L.
+/*
+ *  Copyright © 2011 Igalia S.L.
  *
- *  This program is free software; you can redistribute it and/or modify
+ *  This file is part of Epiphany.
+ *
+ *  Epiphany is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2, or (at your option)
- *  any later version.
+ *  the Free Software Foundation, either version 3 of the License, or
+ *  (at your option) any later version.
  *
- *  This program is distributed in the hope that it will be useful,
+ *  Epiphany is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ *  along with Epiphany.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -281,7 +284,7 @@ ephy_history_service_class_init (EphyHistoryServiceClass *klass)
                          "History filename",
                          "The filename of the SQLite file holding containing history",
                          NULL,
-                         G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE | G_PARAM_STATIC_STRINGS);
+                         G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
 
   obj_properties[PROP_READ_ONLY] =
     g_param_spec_boolean ("read-only",
@@ -949,7 +952,7 @@ ephy_history_service_execute_set_url_thumbnail_time (EphyHistoryService *self,
                                                      EphyHistoryURL     *url,
                                                      gpointer           *result)
 {
-  int thumbnail_time;
+  gint64 thumbnail_time;
 
   if (self->read_only)
     return FALSE;
@@ -968,7 +971,7 @@ ephy_history_service_execute_set_url_thumbnail_time (EphyHistoryService *self,
 void
 ephy_history_service_set_url_thumbnail_time (EphyHistoryService    *self,
                                              const char            *orig_url,
-                                             int                    thumbnail_time,
+                                             gint64                 thumbnail_time,
                                              GCancellable          *cancellable,
                                              EphyHistoryJobCallback callback,
                                              gpointer               user_data)
