@@ -1,22 +1,25 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
+/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /*
  *  Copyright © 2002  Ricardo Fernández Pascual
  *  Copyright © 2003, 2004  Marco Pesenti Gritti
  *  Copyright © 2003, 2004, 2005  Christian Persch
  *  Copyright © 2008  Xan López
+ *  Copyright © 2016  Igalia S.L.
  *
- *  This program is free software; you can redistribute it and/or modify
+ *  This file is part of Epiphany.
+ *
+ *  Epiphany is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2, or (at your option)
- *  any later version.
+ *  the Free Software Foundation, either version 3 of the License, or
+ *  (at your option) any later version.
  *
- *  This program is distributed in the hope that it will be useful,
+ *  Epiphany is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ *  along with Epiphany.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -25,10 +28,10 @@
 #include "ephy-widgets-type-builtins.h"
 #include "ephy-about-handler.h"
 #include "ephy-debug.h"
-#include "ephy-dnd.h"
 #include "ephy-gui.h"
 #include "ephy-lib-type-builtins.h"
 #include "ephy-signal-accumulator.h"
+#include "ephy-title-widget.h"
 #include "ephy-uri-helpers.h"
 
 #include <gdk/gdkkeysyms.h>
@@ -53,13 +56,14 @@
 struct _EphyLocationEntry {
   GtkEntry parent_instance;
 
-  GdkPixbuf *favicon;
   GtkTreeModel *model;
 
   GSList *search_terms;
 
   GBinding *paste_binding;
 
+  GtkPopover *add_bookmark_popover;
+
   char *before_completion;
   char *saved_text;
 
@@ -81,17 +85,8 @@ struct _EphyLocationEntry {
   guint original_address : 1;
   guint apply_colors : 1;
   guint needs_reset : 1;
-  guint show_favicon : 1;
 
-  GtkTargetList *drag_targets;
-  GdkDragAction drag_actions;
-};
-
-static const GtkTargetEntry url_drag_types [] =
-{
-  { (char *)EPHY_DND_URL_TYPE, 0, 0 },
-  { (char *)EPHY_DND_URI_LIST_TYPE, 0, 1 },
-  { (char *)EPHY_DND_TEXT_TYPE, 0, 2 }
+  EphySecurityLevel security_level;
 };
 
 static gboolean ephy_location_entry_reset_internal (EphyLocationEntry *, gboolean);
@@ -104,24 +99,134 @@ static void extracell_data_func (GtkCellLayout   *cell_layout,
 
 enum {
   PROP_0,
-  PROP_LOCATION,
-  PROP_FAVICON,
+  PROP_ADDRESS,
   PROP_SECURITY_LEVEL,
-  PROP_SHOW_FAVICON,
   LAST_PROP
 };
-static GParamSpec *obj_properties[LAST_PROP];
 
 enum signalsEnum {
   USER_CHANGED,
-  LOCK_CLICKED,
+  BOOKMARK_CLICKED,
   GET_LOCATION,
   GET_TITLE,
   LAST_SIGNAL
 };
 static gint signals[LAST_SIGNAL] = { 0 };
 
-G_DEFINE_TYPE (EphyLocationEntry, ephy_location_entry, GTK_TYPE_ENTRY)
+static void ephy_location_entry_title_widget_interface_init (EphyTitleWidgetInterface *iface);
+
+G_DEFINE_TYPE_WITH_CODE (EphyLocationEntry, ephy_location_entry, GTK_TYPE_ENTRY,
+                         G_IMPLEMENT_INTERFACE (EPHY_TYPE_TITLE_WIDGET,
+                                                ephy_location_entry_title_widget_interface_init))
+
+static void
+update_address_state (EphyLocationEntry *entry)
+{
+  const char *text;
+
+  text = gtk_entry_get_text (GTK_ENTRY (entry));
+  entry->original_address = text != NULL &&
+                            g_str_hash (text) == entry->hash;
+}
+
+static const char *
+ephy_location_entry_title_widget_get_address (EphyTitleWidget *widget)
+{
+  EphyLocationEntry *entry = EPHY_LOCATION_ENTRY (widget);
+
+  g_return_val_if_fail (entry, NULL);
+
+  return gtk_entry_get_text (GTK_ENTRY (widget));
+}
+
+static void
+ephy_location_entry_title_widget_set_address (EphyTitleWidget *widget,
+                                              const char      *address)
+{
+  EphyLocationEntry *entry = EPHY_LOCATION_ENTRY (widget);
+  GtkClipboard *clipboard;
+  const char *text;
+  char *effective_text = NULL, *selection = NULL;
+  int start, end;
+
+  g_return_if_fail (widget);
+
+  /* Setting a new text will clear the clipboard. This makes it impossible
+   * to copy&paste from the location entry of one tab into another tab, see
+   * bug #155824. So we save the selection iff the clipboard was owned by
+   * the location entry.
+   */
+  if (gtk_widget_get_realized (GTK_WIDGET (entry))) {
+    clipboard = gtk_widget_get_clipboard (GTK_WIDGET (entry),
+                                          GDK_SELECTION_PRIMARY);
+    g_return_if_fail (clipboard != NULL);
+
+    if (gtk_clipboard_get_owner (clipboard) == G_OBJECT (entry) &&
+        gtk_editable_get_selection_bounds (GTK_EDITABLE (entry),
+                                           &start, &end)) {
+      selection = gtk_editable_get_chars (GTK_EDITABLE (entry),
+                                          start, end);
+    }
+  }
+
+  if (address != NULL) {
+    if (g_str_has_prefix (address, EPHY_ABOUT_SCHEME))
+      effective_text = g_strdup_printf ("about:%s",
+                                        address + strlen (EPHY_ABOUT_SCHEME) + 1);
+    text = address;
+  } else {
+    text = "";
+  }
+
+  /* First record the new hash, then update the entry text */
+  entry->hash = g_str_hash (effective_text ? effective_text : text);
+
+  entry->block_update = TRUE;
+  gtk_entry_set_text (GTK_ENTRY (entry), effective_text ? effective_text : text);
+  entry->block_update = FALSE;
+  g_free (effective_text);
+
+  /* We need to call update_address_state() here, as the 'changed' signal
+   * may not get called if the user has typed in the exact correct url */
+  update_address_state (entry);
+
+  /* Now restore the selection.
+   * Note that it's not owned by the entry anymore!
+   */
+  if (selection != NULL) {
+    gtk_clipboard_set_text (gtk_clipboard_get (GDK_SELECTION_PRIMARY),
+                            selection, strlen (selection));
+    g_free (selection);
+  }
+}
+
+static EphySecurityLevel
+ephy_location_entry_title_widget_get_security_level (EphyTitleWidget *widget)
+{
+  EphyLocationEntry *entry = EPHY_LOCATION_ENTRY (widget);
+
+  g_return_val_if_fail (entry, EPHY_SECURITY_LEVEL_TO_BE_DETERMINED);
+
+  return entry->security_level;
+}
+
+static void
+ephy_location_entry_title_widget_set_security_level (EphyTitleWidget   *widget,
+                                                     EphySecurityLevel  security_level)
+
+{
+  EphyLocationEntry *entry = EPHY_LOCATION_ENTRY (widget);
+  const char *icon_name;
+
+  g_return_if_fail (entry);
+
+  icon_name = ephy_security_level_to_icon_name (security_level);
+  gtk_entry_set_icon_from_icon_name (GTK_ENTRY (widget),
+                                     GTK_ENTRY_ICON_PRIMARY,
+                                     icon_name);
+
+  entry->security_level = security_level;
+}
 
 static void
 ephy_location_entry_set_property (GObject      *object,
@@ -132,21 +237,13 @@ ephy_location_entry_set_property (GObject      *object,
   EphyLocationEntry *entry = EPHY_LOCATION_ENTRY (object);
 
   switch (prop_id) {
-    case PROP_LOCATION:
-      ephy_location_entry_set_location (entry,
-                                        g_value_get_string (value));
-      break;
-    case PROP_FAVICON:
-      ephy_location_entry_set_favicon (entry,
-                                       g_value_get_object (value));
+    case PROP_ADDRESS:
+      ephy_title_widget_set_address (EPHY_TITLE_WIDGET (entry),
+                                     g_value_get_string (value));
       break;
     case PROP_SECURITY_LEVEL:
-      ephy_location_entry_set_security_level (entry,
-                                              g_value_get_enum (value));
-      break;
-    case PROP_SHOW_FAVICON:
-      ephy_location_entry_set_show_favicon (entry,
-                                            g_value_get_boolean (value));
+      ephy_title_widget_set_security_level (EPHY_TITLE_WIDGET (entry),
+                                            g_value_get_enum (value));
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -162,8 +259,11 @@ ephy_location_entry_get_property (GObject    *object,
   EphyLocationEntry *entry = EPHY_LOCATION_ENTRY (object);
 
   switch (prop_id) {
-    case PROP_LOCATION:
-      g_value_set_string (value, ephy_location_entry_get_location (entry));
+    case PROP_ADDRESS:
+      g_value_set_string (value, ephy_title_widget_get_address (EPHY_TITLE_WIDGET (entry)));
+      break;
+    case PROP_SECURITY_LEVEL:
+      g_value_set_enum (value, ephy_title_widget_get_security_level (EPHY_TITLE_WIDGET (entry)));
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -177,17 +277,31 @@ ephy_location_entry_finalize (GObject *object)
 
   g_free (entry->saved_text);
 
-  if (entry->drag_targets != NULL) {
-    gtk_target_list_unref (entry->drag_targets);
-  }
-
-  if (entry->favicon != NULL) {
-    g_object_unref (entry->favicon);
-  }
-
   G_OBJECT_CLASS (ephy_location_entry_parent_class)->finalize (object);
 }
 
+static void
+ephy_location_entry_size_allocate (GtkWidget     *widget,
+                                   GtkAllocation *allocation)
+{
+  EphyLocationEntry *entry = EPHY_LOCATION_ENTRY (widget);
+
+  if (gtk_widget_is_visible (widget)) {
+    GdkRectangle pointing_to;
+
+    gtk_entry_get_icon_area (GTK_ENTRY (entry),
+                             GTK_ENTRY_ICON_SECONDARY,
+                             &pointing_to);
+    /* FIXME: GTK+ sets "margin-left: 6px" for the icon. Add 3px to the
+     * rectangle so the popover is centered on the star.
+     */
+    pointing_to.x = pointing_to.x + 3;
+    gtk_popover_set_pointing_to (entry->add_bookmark_popover, &pointing_to);
+  }
+
+  GTK_WIDGET_CLASS (ephy_location_entry_parent_class)->size_allocate (widget, allocation);
+}
+
 static void
 ephy_location_entry_get_preferred_width (GtkWidget *widget,
                                          gint      *minimum_width,
@@ -236,6 +350,15 @@ ephy_location_entry_cut_clipboard (GtkEntry *entry)
   gtk_editable_delete_selection (GTK_EDITABLE (entry));
 }
 
+static void
+ephy_location_entry_title_widget_interface_init (EphyTitleWidgetInterface *iface)
+{
+  iface->get_address = ephy_location_entry_title_widget_get_address;
+  iface->set_address = ephy_location_entry_title_widget_set_address;
+  iface->get_security_level = ephy_location_entry_title_widget_get_security_level;
+  iface->set_security_level = ephy_location_entry_title_widget_set_security_level;
+}
+
 static void
 ephy_location_entry_class_init (EphyLocationEntryClass *klass)
 {
@@ -246,55 +369,15 @@ ephy_location_entry_class_init (EphyLocationEntryClass *klass)
   object_class->get_property = ephy_location_entry_get_property;
   object_class->set_property = ephy_location_entry_set_property;
   object_class->finalize = ephy_location_entry_finalize;
+
+  widget_class->size_allocate = ephy_location_entry_size_allocate;
   widget_class->get_preferred_width = ephy_location_entry_get_preferred_width;
+
   entry_class->copy_clipboard = ephy_location_entry_copy_clipboard;
   entry_class->cut_clipboard = ephy_location_entry_cut_clipboard;
 
-  /**
-   * EphyLocationEntry:location:
-   *
-   * The current location.
-   */
-  obj_properties[PROP_LOCATION] =
-    g_param_spec_string ("location",
-                         "Location",
-                         "The current location",
-                         "",
-                         G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
-
-  /**
-   * EphyLocationEntry:favicon:
-   *
-   * The icon corresponding to the current location.
-   */
-  obj_properties[PROP_FAVICON] =
-    g_param_spec_object ("favicon",
-                         "Favicon",
-                         "The icon corresponding to the current location",
-                         GDK_TYPE_PIXBUF,
-                         G_PARAM_WRITABLE | G_PARAM_STATIC_STRINGS);
-
-  /**
-   * EphyLocationEntry:security-level:
-   *
-   * State of the security icon.
-   */
-  obj_properties[PROP_SECURITY_LEVEL] =
-    g_param_spec_enum ("security-level",
-                       "Security level",
-                       "State of the security icon",
-                       EPHY_TYPE_SECURITY_LEVEL,
-                       EPHY_SECURITY_LEVEL_TO_BE_DETERMINED,
-                       G_PARAM_WRITABLE | G_PARAM_STATIC_STRINGS);
-
-  obj_properties[PROP_SHOW_FAVICON] =
-    g_param_spec_boolean ("show-favicon",
-                          "Show Favicon",
-                          "Whether to show the favicon",
-                          TRUE,
-                          G_PARAM_WRITABLE | G_PARAM_STATIC_STRINGS);
-
-  g_object_class_install_properties (object_class, LAST_PROP, obj_properties);
+  g_object_class_override_property (object_class, PROP_ADDRESS, "address");
+  g_object_class_override_property (object_class, PROP_SECURITY_LEVEL, "security-level");
 
   /**
    * EphyLocationEntry::user-changed:
@@ -303,29 +386,26 @@ ephy_location_entry_class_init (EphyLocationEntryClass *klass)
    * Emitted when the user changes the contents of the internal #GtkEntry
    *
    */
-  signals[USER_CHANGED] = g_signal_new (
-    "user_changed", G_OBJECT_CLASS_TYPE (klass),
-    G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST,
-    0, NULL, NULL, NULL,
-    G_TYPE_NONE,
-    0,
-    G_TYPE_NONE);
+  signals[USER_CHANGED] = g_signal_new ("user_changed", G_OBJECT_CLASS_TYPE (klass),
+                                        G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST,
+                                        0, NULL, NULL, NULL,
+                                        G_TYPE_NONE,
+                                        0,
+                                        G_TYPE_NONE);
 
   /**
-   * EphyLocationEntry::lock-clicked:
+   * EphyLocationEntry::bookmark-clicked:
    * @entry: the object on which the signal is emitted
    *
-   * Emitted when the user clicks the security icon inside the
+   * Emitted when the user clicks the bookmark icon inside the
    * #EphyLocationEntry.
    *
    */
-  signals[LOCK_CLICKED] = g_signal_new (
-    "lock-clicked",
-    EPHY_TYPE_LOCATION_ENTRY,
-    G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST,
-    0, NULL, NULL, NULL,
-    G_TYPE_NONE,
-    0);
+  signals[BOOKMARK_CLICKED] = g_signal_new ("bookmark-clicked", G_OBJECT_CLASS_TYPE (klass),
+                                            G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST,
+                                            0, NULL, NULL, NULL,
+                                            G_TYPE_NONE,
+                                            0);
 
   /**
    * EphyLocationEntry::get-location:
@@ -337,14 +417,13 @@ ephy_location_entry_class_init (EphyLocationEntryClass *klass)
    * function should return the address of the currently loaded site.
    *
    */
-  signals[GET_LOCATION] = g_signal_new (
-    "get-location", G_OBJECT_CLASS_TYPE (klass),
-    G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST,
-    0, ephy_signal_accumulator_string,
-    NULL, NULL,
-    G_TYPE_STRING,
-    0,
-    G_TYPE_NONE);
+  signals[GET_LOCATION] = g_signal_new ("get-location", G_OBJECT_CLASS_TYPE (klass),
+                                        G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST,
+                                        0, ephy_signal_accumulator_string,
+                                        NULL, NULL,
+                                        G_TYPE_STRING,
+                                        0,
+                                        G_TYPE_NONE);
 
   /**
    * EphyLocationEntry::get-title:
@@ -356,58 +435,13 @@ ephy_location_entry_class_init (EphyLocationEntryClass *klass)
    * function should return the title of the currently loaded site.
    *
    */
-  signals[GET_TITLE] = g_signal_new (
-    "get-title", G_OBJECT_CLASS_TYPE (klass),
-    G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST,
-    0, ephy_signal_accumulator_string,
-    NULL, NULL,
-    G_TYPE_STRING,
-    0,
-    G_TYPE_NONE);
-}
-
-static void
-update_address_state (EphyLocationEntry *entry)
-{
-  const char *text;
-
-  text = gtk_entry_get_text (GTK_ENTRY (entry));
-  entry->original_address = text != NULL &&
-                            g_str_hash (text) == entry->hash;
-}
-
-static void
-update_favicon (EphyLocationEntry *lentry)
-{
-  GtkEntry *entry = GTK_ENTRY (lentry);
-
-  /* Only show the favicon if the entry's text is the
-   * address of the current page.
-   */
-  if (lentry->show_favicon && lentry->favicon != NULL && lentry->original_address) {
-    gtk_entry_set_icon_from_pixbuf (entry,
-                                    GTK_ENTRY_ICON_PRIMARY,
-                                    lentry->favicon);
-  } else if (lentry->show_favicon) {
-    const char *icon_name;
-
-    /* Here we could consider using fallback favicon that matches
-     * the page MIME type, though text/html should be good enough
-     * most of the time. See #337140
-     */
-    if (gtk_entry_get_text_length (entry) > 0)
-      icon_name = "text-x-generic-symbolic";
-    else
-      icon_name = "edit-find-symbolic";
-
-    gtk_entry_set_icon_from_icon_name (entry,
-                                       GTK_ENTRY_ICON_PRIMARY,
-                                       icon_name);
-  } else {
-    gtk_entry_set_icon_from_icon_name (entry,
-                                       GTK_ENTRY_ICON_PRIMARY,
-                                       NULL);
-  }
+  signals[GET_TITLE] = g_signal_new ("get-title", G_OBJECT_CLASS_TYPE (klass),
+                                     G_SIGNAL_RUN_FIRST | G_SIGNAL_RUN_LAST,
+                                     0, ephy_signal_accumulator_string,
+                                     NULL, NULL,
+                                     G_TYPE_STRING,
+                                     0,
+                                     G_TYPE_NONE);
 }
 
 static void
@@ -515,7 +549,7 @@ match_selected_cb (GtkEntryCompletion *completion,
   entry->needs_reset = (state == GDK_CONTROL_MASK ||
                         state == (GDK_CONTROL_MASK | GDK_SHIFT_MASK));
 
-  ephy_location_entry_set_location (entry, item);
+  ephy_title_widget_set_address (EPHY_TITLE_WIDGET (entry), item);
   /* gtk_im_context_reset (GTK_ENTRY (entry)->im_context); */
   g_signal_emit_by_name (entry, "activate");
 
@@ -539,26 +573,6 @@ action_activated_after_cb (GtkEntryCompletion *completion,
   }
 }
 
-static gboolean
-entry_drag_motion_cb (GtkWidget      *widget,
-                      GdkDragContext *context,
-                      gint            x,
-                      gint            y,
-                      guint           time)
-{
-  return FALSE;
-}
-
-static gboolean
-entry_drag_drop_cb (GtkWidget      *widget,
-                    GdkDragContext *context,
-                    gint            x,
-                    gint            y,
-                    guint           time)
-{
-  return FALSE;
-}
-
 static void
 entry_clear_activate_cb (GtkMenuItem       *item,
                          EphyLocationEntry *entry)
@@ -606,6 +620,13 @@ entry_undo_activate_cb (GtkMenuItem       *item,
   ephy_location_entry_reset_internal (entry, FALSE);
 }
 
+/* The build should fail here each time when upgrading to a new major version
+ * of GTK+, so that we don't forget to update this domain.
+ */
+#if GTK_MAJOR_VERSION == 3
+#define GTK_GETTEXT_DOMAIN "gtk30"
+#endif
+
 static void
 entry_populate_popup_cb (GtkEntry          *entry,
                          GtkMenu           *menu,
@@ -616,7 +637,7 @@ entry_populate_popup_cb (GtkEntry          *entry,
   GtkWidget *redo_menuitem;
   GtkWidget *paste_and_go_menuitem;
   GtkWidget *separator;
-  GtkWidget *paste_menuitem;
+  GtkWidget *paste_menuitem = NULL;
   GList *children, *item;
   int pos = 0, sep = 0;
   gboolean is_editable;
@@ -649,11 +670,12 @@ entry_populate_popup_cb (GtkEntry          *entry,
   /* Search for the Paste menu item and insert right after it. */
   children = gtk_container_get_children (GTK_CONTAINER (menu));
   for (item = children, pos = 0; item != NULL; item = item->next, pos++) {
-    if (g_strcmp0 (gtk_menu_item_get_label (item->data), _("_Paste")) == 0) {
+    if (g_strcmp0 (gtk_menu_item_get_label (item->data), g_dgettext (GTK_GETTEXT_DOMAIN, "_Paste")) == 0) {
       paste_menuitem = item->data;
       break;
     }
   }
+  g_assert (paste_menuitem != NULL);
 
   g_signal_connect (paste_and_go_menuitem, "activate",
                     G_CALLBACK (entry_paste_and_go_activate_cb), lentry);
@@ -682,217 +704,6 @@ entry_populate_popup_cb (GtkEntry          *entry,
   gtk_menu_shell_insert (GTK_MENU_SHELL (menu), separator, 2);
 }
 
-static void
-each_url_get_data_binder (EphyDragEachSelectedItemDataGet iteratee,
-                          gpointer                        iterator_context,
-                          gpointer                        return_data)
-{
-  EphyLocationEntry *entry = EPHY_LOCATION_ENTRY (iterator_context);
-  char *title = NULL, *address = NULL;
-
-  g_signal_emit (entry, signals[GET_LOCATION], 0, &address);
-  g_signal_emit (entry, signals[GET_TITLE], 0, &title);
-  g_return_if_fail (address != NULL && title != NULL);
-
-  iteratee (address, title, return_data);
-
-  g_free (address);
-  g_free (title);
-}
-
-static void
-sanitize_location (char **url)
-{
-  char *str;
-
-  /* Do not show internal ephy-about: protocol to users */
-  if (g_str_has_prefix (*url, EPHY_ABOUT_SCHEME)) {
-    str = g_strdup_printf ("about:%s", *url + strlen (EPHY_ABOUT_SCHEME) + 1);
-    g_free (*url);
-    *url = str;
-  }
-}
-
-#define DRAG_ICON_LAYOUT_PADDING        5
-#define DRAG_ICON_ICON_PADDING          10
-#define DRAG_ICON_MAX_WIDTH_CHARS       32
-
-static cairo_surface_t *
-favicon_create_drag_surface (EphyLocationEntry *entry,
-                             GtkWidget         *widget)
-{
-  char *title = NULL, *address = NULL;
-  GString *text;
-  GtkStyleContext *style;
-  const PangoFontDescription *font_desc;
-  cairo_surface_t *surface;
-  PangoContext *context;
-  PangoLayout  *layout;
-  PangoFontMetrics *metrics;
-  int surface_height, surface_width;
-  int layout_width, layout_height;
-  int icon_width = 0, icon_height = 0, favicon_offset_x = 0;
-  int char_width;
-  cairo_t *cr;
-  GtkStateFlags state;
-  GdkRGBA color;
-  GdkPixbuf *favicon;
-
-  g_signal_emit (entry, signals[GET_LOCATION], 0, &address);
-  sanitize_location (&address);
-  g_signal_emit (entry, signals[GET_TITLE], 0, &title);
-  if (address == NULL || title == NULL) return NULL;
-
-  /* Compute text */
-  title = g_strstrip (title);
-
-  text = g_string_sized_new (strlen (address) + strlen (title) + 2);
-  if (title[0] != '\0') {
-    g_string_append (text, title);
-    g_string_append (text, "\n");
-  }
-
-  if (address[0] != '\0') {
-    g_string_append (text, address);
-  }
-
-  if (entry->favicon != NULL)
-    favicon = g_object_ref (entry->favicon);
-  else
-    favicon = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
-                                        "text-x-generic-symbolic",
-                                        16,
-                                        0, NULL);
-  if (favicon != NULL) {
-    icon_width = gdk_pixbuf_get_width (favicon);
-    icon_height = gdk_pixbuf_get_height (favicon);
-  }
-
-  context = gtk_widget_get_pango_context (widget);
-  layout = pango_layout_new (context);
-
-  style = gtk_widget_get_style_context (GTK_WIDGET (entry));
-  state = gtk_style_context_get_state (style);
-
-  gtk_style_context_save (style);
-  gtk_style_context_set_state (style, GTK_STATE_FLAG_NORMAL);
-  gtk_style_context_get (style, GTK_STATE_FLAG_NORMAL,
-                         "font", &font_desc, NULL);
-  gtk_style_context_restore (style);
-
-  metrics = pango_context_get_metrics (context,
-                                       font_desc,
-                                       pango_context_get_language (context));
-
-  char_width = pango_font_metrics_get_approximate_digit_width (metrics);
-  pango_font_metrics_unref (metrics);
-
-  pango_layout_set_ellipsize (layout, PANGO_ELLIPSIZE_END);
-  pango_layout_set_width (layout, char_width * DRAG_ICON_MAX_WIDTH_CHARS);
-  pango_layout_set_text (layout, text->str, text->len);
-
-  pango_layout_get_pixel_size (layout, &layout_width, &layout_height);
-
-  if (favicon != NULL) {
-    favicon_offset_x = icon_width + (2 * DRAG_ICON_ICON_PADDING);
-  }
-
-  surface_width = layout_width + favicon_offset_x +
-                  (DRAG_ICON_LAYOUT_PADDING * 3);
-  surface_height = MAX (layout_height, icon_height) +
-                   (DRAG_ICON_LAYOUT_PADDING * 2);
-
-  surface = gdk_window_create_similar_surface (gtk_widget_get_window (widget),
-                                               CAIRO_CONTENT_COLOR,
-                                               surface_width + 2,
-                                               surface_height + 2);
-  cr = cairo_create (surface);
-
-  cairo_rectangle (cr, 1, 1, surface_width, surface_height);
-  cairo_set_line_width (cr, 1.0);
-
-  cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
-  cairo_stroke_preserve (cr);
-
-  gtk_style_context_get_background_color (style, state, &color);
-  gdk_cairo_set_source_rgba (cr, &color);
-  cairo_fill (cr);
-
-  if (favicon != NULL) {
-    double x;
-    double y;
-
-    x = 1 + DRAG_ICON_LAYOUT_PADDING + DRAG_ICON_ICON_PADDING;
-    y = (surface_height - icon_height) / 2;
-    gdk_cairo_set_source_pixbuf (cr, favicon, x, y);
-    cairo_rectangle (cr, x, y, icon_width, icon_height);
-    cairo_fill (cr);
-  }
-
-  cairo_move_to (cr,
-                 1 + DRAG_ICON_LAYOUT_PADDING + favicon_offset_x,
-                 1 + DRAG_ICON_LAYOUT_PADDING);
-  gtk_style_context_get_color (style, state, &color);
-  gdk_cairo_set_source_rgba (cr, &color);
-  pango_cairo_show_layout (cr, layout);
-
-  cairo_destroy (cr);
-  g_object_unref (layout);
-
-  g_free (address);
-  g_free (title);
-  g_string_free (text, TRUE);
-  g_clear_object (&favicon);
-
-  return surface;
-}
-
-static void
-favicon_drag_begin_cb (GtkWidget         *widget,
-                       GdkDragContext    *context,
-                       EphyLocationEntry *lentry)
-{
-  cairo_surface_t *surface;
-  GtkEntry *entry;
-  gint index;
-
-  entry = GTK_ENTRY (widget);
-
-  index = gtk_entry_get_current_icon_drag_source (entry);
-  if (index != GTK_ENTRY_ICON_PRIMARY)
-    return;
-
-  surface = favicon_create_drag_surface (lentry, widget);
-
-  if (surface != NULL) {
-    gtk_drag_set_icon_surface (context, surface);
-    cairo_surface_destroy (surface);
-  }
-}
-
-static void
-favicon_drag_data_get_cb (GtkWidget         *widget,
-                          GdkDragContext    *context,
-                          GtkSelectionData  *selection_data,
-                          guint              info,
-                          guint32            time,
-                          EphyLocationEntry *lentry)
-{
-  gint index;
-  GtkEntry *entry;
-
-  g_assert (widget != NULL);
-  g_return_if_fail (context != NULL);
-
-  entry = GTK_ENTRY (widget);
-
-  index = gtk_entry_get_current_icon_drag_source (entry);
-  if (index == GTK_ENTRY_ICON_PRIMARY) {
-    ephy_dnd_drag_data_get (widget, context, selection_data,
-                            time, lentry, each_url_get_data_binder);
-  }
-}
-
 static gboolean
 icon_button_press_event_cb (GtkWidget           *entry,
                             GtkEntryIconPosition position,
@@ -905,16 +716,12 @@ icon_button_press_event_cb (GtkWidget           *entry,
       event->button == 1 &&
       state == 0 /* left */) {
     if (position == GTK_ENTRY_ICON_PRIMARY) {
-      GtkWidget *toplevel;
-
-      toplevel = gtk_widget_get_toplevel (GTK_WIDGET (entry));
-      gtk_window_set_focus (GTK_WINDOW (toplevel), entry);
-
-      gtk_editable_select_region (GTK_EDITABLE (entry), 0, -1);
+      GdkRectangle lock_position;
+      gtk_entry_get_icon_area (GTK_ENTRY (entry), GTK_ENTRY_ICON_PRIMARY, &lock_position);
+      g_signal_emit_by_name (entry, "lock-clicked", &lock_position);
     } else {
-      g_signal_emit (lentry, signals[LOCK_CLICKED], 0);
+      g_signal_emit (entry, signals[BOOKMARK_CLICKED], 0);
     }
-
     return TRUE;
   }
 
@@ -928,42 +735,21 @@ ephy_location_entry_construct_contents (EphyLocationEntry *lentry)
 
   LOG ("EphyLocationEntry constructing contents %p", lentry);
 
-  /* Favicon */
-  lentry->drag_targets = gtk_target_list_new (url_drag_types,
-                                              G_N_ELEMENTS (url_drag_types));
-  lentry->drag_actions = GDK_ACTION_ASK | GDK_ACTION_COPY | GDK_ACTION_LINK;
-
-  gtk_entry_set_icon_drag_source (GTK_ENTRY (entry),
-                                  GTK_ENTRY_ICON_PRIMARY,
-                                  lentry->drag_targets,
-                                  lentry->drag_actions);
-
-  gtk_entry_set_icon_tooltip_text (GTK_ENTRY (entry),
-                                   GTK_ENTRY_ICON_PRIMARY,
-                                   _("Drag and drop this icon to create a link to this page"));
-
-  gtk_drag_dest_set (entry,
-                     GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_DROP,
-                     url_drag_types,
-                     G_N_ELEMENTS (url_drag_types),
-                     GDK_ACTION_MOVE | GDK_ACTION_COPY);
+  gtk_entry_set_icon_from_icon_name (GTK_ENTRY (entry),
+                                     GTK_ENTRY_ICON_SECONDARY,
+                                     "non-starred-symbolic");
 
   g_object_connect (entry,
                     "signal::icon-press", G_CALLBACK (icon_button_press_event_cb), lentry,
                     "signal::populate-popup", G_CALLBACK (entry_populate_popup_cb), lentry,
                     "signal::key-press-event", G_CALLBACK (entry_key_press_cb), lentry,
                     "signal::changed", G_CALLBACK (editable_changed_cb), lentry,
-                    "signal::drag-motion", G_CALLBACK (entry_drag_motion_cb), lentry,
-                    "signal::drag-drop", G_CALLBACK (entry_drag_drop_cb), lentry,
-                    "signal::drag-data-get", G_CALLBACK (favicon_drag_data_get_cb), lentry,
                     NULL);
 
   g_signal_connect_after (entry, "key-press-event",
                           G_CALLBACK (entry_key_press_after_cb), lentry);
   g_signal_connect_after (entry, "activate",
                           G_CALLBACK (entry_activate_after_cb), lentry);
-  g_signal_connect_after (entry, "drag-begin",
-                          G_CALLBACK (favicon_drag_begin_cb), lentry);
 }
 
 static void
@@ -974,7 +760,6 @@ ephy_location_entry_init (EphyLocationEntry *le)
   le->user_changed = FALSE;
   le->block_update = FALSE;
   le->saved_text = NULL;
-  le->show_favicon = TRUE;
   le->dns_prefetch_handler = 0;
 
   ephy_location_entry_construct_contents (le);
@@ -1273,81 +1058,6 @@ ephy_location_entry_set_completion (EphyLocationEntry *entry,
   g_object_unref (completion);
 }
 
-/**
- * ephy_location_entry_set_location:
- * @entry: an #EphyLocationEntry widget
- * @address: new location address
- *
- * Sets the current address of @entry to @address.
- **/
-void
-ephy_location_entry_set_location (EphyLocationEntry *entry,
-                                  const char        *address)
-{
-  GtkWidget *widget = GTK_WIDGET (entry);
-  GtkClipboard *clipboard;
-  const char *text;
-  char *effective_text = NULL, *selection = NULL;
-  int start, end;
-
-  /* Setting a new text will clear the clipboard. This makes it impossible
-   * to copy&paste from the location entry of one tab into another tab, see
-   * bug #155824. So we save the selection iff the clipboard was owned by
-   * the location entry.
-   */
-  if (gtk_widget_get_realized (widget)) {
-    clipboard = gtk_widget_get_clipboard (widget,
-                                          GDK_SELECTION_PRIMARY);
-    g_return_if_fail (clipboard != NULL);
-
-    if (gtk_clipboard_get_owner (clipboard) == G_OBJECT (widget) &&
-        gtk_editable_get_selection_bounds (GTK_EDITABLE (widget),
-                                           &start, &end)) {
-      selection = gtk_editable_get_chars (GTK_EDITABLE (widget),
-                                          start, end);
-    }
-  }
-
-  if (address != NULL) {
-    if (g_str_has_prefix (address, EPHY_ABOUT_SCHEME))
-      effective_text = g_strdup_printf ("about:%s",
-                                        address + strlen (EPHY_ABOUT_SCHEME) + 1);
-    text = address;
-    gtk_entry_set_icon_drag_source (GTK_ENTRY (entry),
-                                    GTK_ENTRY_ICON_PRIMARY,
-                                    entry->drag_targets,
-                                    entry->drag_actions);
-  } else {
-    text = "";
-    gtk_entry_set_icon_drag_source (GTK_ENTRY (entry),
-                                    GTK_ENTRY_ICON_PRIMARY,
-                                    NULL,
-                                    GDK_ACTION_DEFAULT);
-  }
-
-  /* First record the new hash, then update the entry text */
-  entry->hash = g_str_hash (effective_text ? effective_text : text);
-
-  entry->block_update = TRUE;
-  gtk_entry_set_text (GTK_ENTRY (widget), effective_text ? effective_text : text);
-  entry->block_update = FALSE;
-  g_free (effective_text);
-
-  /* We need to call update_address_state() here, as the 'changed' signal
-   * may not get called if the user has typed in the exact correct url */
-  update_address_state (entry);
-  update_favicon (entry);
-
-  /* Now restore the selection.
-   * Note that it's not owned by the entry anymore!
-   */
-  if (selection != NULL) {
-    gtk_clipboard_set_text (gtk_clipboard_get (GDK_SELECTION_PRIMARY),
-                            selection, strlen (selection));
-    g_free (selection);
-  }
-}
-
 /**
  * ephy_location_entry_get_can_undo:
  * @entry: an #EphyLocationEntry widget
@@ -1380,22 +1090,6 @@ ephy_location_entry_get_can_redo (EphyLocationEntry *entry)
   return entry->can_redo;
 }
 
-/**
- * ephy_location_entry_get_location:
- * @entry: an #EphyLocationEntry widget
- *
- * Retrieves the text displayed by the internal #GtkEntry of @entry. This is
- * the currently displayed text, like in any #GtkEntry.
- *
- * Return value: the text inside the inner #GtkEntry of @entry, owned by GTK+
- *
- **/
-const char *
-ephy_location_entry_get_location (EphyLocationEntry *entry)
-{
-  return gtk_entry_get_text (GTK_ENTRY (entry));
-}
-
 static gboolean
 ephy_location_entry_reset_internal (EphyLocationEntry *entry,
                                     gboolean           notify)
@@ -1415,7 +1109,7 @@ ephy_location_entry_reset_internal (EphyLocationEntry *entry,
 
   retval = g_str_hash (text) != g_str_hash (old_text);
 
-  ephy_location_entry_set_location (entry, text);
+  ephy_title_widget_set_address (EPHY_TITLE_WIDGET (entry), text);
   g_free (url);
 
   if (notify) {
@@ -1480,59 +1174,41 @@ ephy_location_entry_activate (EphyLocationEntry *entry)
                         widget);
 }
 
-/**
- * ephy_location_entry_set_favicon:
- * @entry: an #EphyLocationEntry widget
- * @pixbuf: a #GdkPixbuf to be set as the icon of the entry
- *
- * Sets the icon in the internal #EphyIconEntry of @entry
- *
- **/
 void
-ephy_location_entry_set_favicon (EphyLocationEntry *entry,
-                                 GdkPixbuf         *pixbuf)
+ephy_location_entry_set_bookmark_icon_state (EphyLocationEntry                  *entry,
+                                             EphyLocationEntryBookmarkIconState  state)
 {
-  if (entry->favicon != NULL) {
-    g_object_unref (entry->favicon);
+  GtkStyleContext *context;
+
+  g_return_if_fail (EPHY_IS_LOCATION_ENTRY (entry));
+
+  context = gtk_widget_get_style_context (GTK_WIDGET (entry));
+
+  switch (state) {
+    case EPHY_LOCATION_ENTRY_BOOKMARK_ICON_HIDDEN:
+      gtk_entry_set_icon_from_icon_name (GTK_ENTRY (entry),
+                                         GTK_ENTRY_ICON_SECONDARY,
+                                         NULL);
+      gtk_style_context_remove_class (context, "starred");
+      gtk_style_context_remove_class (context, "non-starred");
+      break;
+    case EPHY_LOCATION_ENTRY_BOOKMARK_ICON_EMPTY:
+      gtk_entry_set_icon_from_icon_name (GTK_ENTRY (entry),
+                                         GTK_ENTRY_ICON_SECONDARY,
+                                         "non-starred-symbolic");
+      gtk_style_context_remove_class (context, "starred");
+      gtk_style_context_add_class (context, "non-starred");
+      break;
+    case EPHY_LOCATION_ENTRY_BOOKMARK_ICON_BOOKMARKED:
+      gtk_entry_set_icon_from_icon_name (GTK_ENTRY (entry),
+                                         GTK_ENTRY_ICON_SECONDARY,
+                                         "starred-symbolic");
+      gtk_style_context_remove_class (context, "non-starred");
+      gtk_style_context_add_class (context, "starred");
+      break;
+    default:
+      g_assert_not_reached ();
   }
-
-  entry->favicon = pixbuf ? g_object_ref (pixbuf) : NULL;
-
-  update_favicon (entry);
-}
-
-void
-ephy_location_entry_set_show_favicon (EphyLocationEntry *entry,
-                                      gboolean           show_favicon)
-{
-  g_return_if_fail (EPHY_IS_LOCATION_ENTRY (entry));
-
-  entry->show_favicon = show_favicon != FALSE;
-
-  update_favicon (entry);
-}
-
-/**
- * ephy_location_entry_set_security_level:
- * @entry: an #EphyLocationEntry widget
- * @state: the #EphySecurityLevel
- *
- * Set the lock icon to be displayed
- *
- **/
-void
-ephy_location_entry_set_security_level (EphyLocationEntry *entry,
-                                        EphySecurityLevel  security_level)
-
-{
-  const char *icon_name;
-
-  g_return_if_fail (EPHY_IS_LOCATION_ENTRY (entry));
-
-  icon_name = ephy_security_level_to_icon_name (security_level);
-  gtk_entry_set_icon_from_icon_name (GTK_ENTRY (entry),
-                                     GTK_ENTRY_ICON_SECONDARY,
-                                     icon_name);
 }
 
 /**
@@ -1548,10 +1224,26 @@ ephy_location_entry_set_lock_tooltip (EphyLocationEntry *entry,
                                       const char        *tooltip)
 {
   gtk_entry_set_icon_tooltip_text (GTK_ENTRY (entry),
-                                   GTK_ENTRY_ICON_SECONDARY,
+                                   GTK_ENTRY_ICON_PRIMARY,
                                    tooltip);
 }
 
+void
+ephy_location_entry_set_add_bookmark_popover (EphyLocationEntry *entry,
+                                              GtkPopover        *popover)
+{
+  g_return_if_fail (EPHY_IS_LOCATION_ENTRY (entry));
+  g_return_if_fail (GTK_IS_POPOVER (popover));
+
+  entry->add_bookmark_popover = popover;
+}
+
+GtkPopover *
+ephy_location_entry_get_add_bookmark_popover (EphyLocationEntry *entry)
+{
+  return entry->add_bookmark_popover;
+}
+
 /**
  * ephy_location_entry_get_search_terms:
  * @entry: an #EphyLocationEntry widget
