@@ -1,21 +1,23 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
+/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
 /*
  *  Copyright © 2002 Christophe Fergeau
  *  Copyright © 2003, 2004 Marco Pesenti Gritti
  *  Copyright © 2003, 2004, 2005 Christian Persch
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2, or (at your option)
- *  any later version.
+ *  This file is part of Epiphany.
  *
- *  This program is distributed in the hope that it will be useful,
+ *  Epiphany is free software: you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation, either version 3 of the License, or
+ *  (at your option) any later version.
+ *
+ *  Epiphany is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ *  along with Epiphany.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -38,7 +40,6 @@
 #define TAB_WIDTH_N_CHARS 15
 
 #define AFTER_ALL_TABS -1
-#define NOT_IN_APP_WINDOWS -2
 
 #define INSANE_NUMBER_OF_URLS 20
 
@@ -50,9 +51,12 @@ struct _EphyNotebook {
   GList *focused_pages;
   guint tabs_vis_notifier_id;
 
+  GMenu *tab_menu;
+
   guint tabs_allowed : 1;
 };
 
+static void ephy_notebook_constructed (GObject *object);
 static void ephy_notebook_finalize (GObject *object);
 static int  ephy_notebook_insert_page (GtkNotebook *notebook,
                                        GtkWidget   *child,
@@ -61,6 +65,15 @@ static int  ephy_notebook_insert_page (GtkNotebook *notebook,
                                        int          position);
 static void ephy_notebook_remove (GtkContainer *container,
                                   GtkWidget    *tab_widget);
+static void ephy_notebook_page_added (GtkNotebook *notebook,
+                                      GtkWidget   *child,
+                                      guint        page_num);
+static void ephy_notebook_page_removed (GtkNotebook *notebook,
+                                        GtkWidget   *child,
+                                        guint        page_num);
+static void ephy_notebook_page_reordered (GtkNotebook *notebook,
+                                          GtkWidget   *child,
+                                          guint        page_num);
 
 static const GtkTargetEntry url_drag_types [] =
 {
@@ -129,6 +142,7 @@ ephy_notebook_class_init (EphyNotebookClass *klass)
   GtkContainerClass *container_class = GTK_CONTAINER_CLASS (klass);
   GtkNotebookClass *notebook_class = GTK_NOTEBOOK_CLASS (klass);
 
+  object_class->constructed = ephy_notebook_constructed;
   object_class->finalize = ephy_notebook_finalize;
   object_class->get_property = ephy_notebook_get_property;
   object_class->set_property = ephy_notebook_set_property;
@@ -136,6 +150,9 @@ ephy_notebook_class_init (EphyNotebookClass *klass)
   container_class->remove = ephy_notebook_remove;
 
   notebook_class->insert_page = ephy_notebook_insert_page;
+  notebook_class->page_added = ephy_notebook_page_added;
+  notebook_class->page_removed = ephy_notebook_page_removed;
+  notebook_class->page_reordered = ephy_notebook_page_reordered;
 
   signals[TAB_CLOSE_REQUEST] =
     g_signal_new ("tab-close-request",
@@ -156,50 +173,6 @@ ephy_notebook_class_init (EphyNotebookClass *klass)
   g_object_class_install_properties (object_class, LAST_PROP, obj_properties);
 }
 
-
-/* FIXME remove when gtknotebook's func for this becomes public, bug #.... */
-static EphyNotebook *
-find_notebook_at_pointer (GdkDisplay *display, gint abs_x, gint abs_y)
-{
-  GdkWindow *win_at_pointer, *toplevel_win;
-  gpointer toplevel = NULL;
-  gint x, y;
-
-  win_at_pointer = gdk_device_get_window_at_position (
-    gdk_device_manager_get_client_pointer (
-      gdk_display_get_device_manager (display)),
-    &x, &y);
-  if (win_at_pointer == NULL) {
-    /* We are outside all windows containing a notebook */
-    return NULL;
-  }
-
-  toplevel_win = gdk_window_get_toplevel (win_at_pointer);
-
-  /* get the GtkWidget which owns the toplevel GdkWindow */
-  gdk_window_get_user_data (toplevel_win, &toplevel);
-
-  /* toplevel should be an EphyWindow */
-  if (toplevel != NULL && EPHY_IS_WINDOW (toplevel)) {
-    return EPHY_NOTEBOOK (ephy_window_get_notebook
-                            (EPHY_WINDOW (toplevel)));
-  }
-
-  return NULL;
-}
-
-static gboolean
-is_in_notebook_window (EphyNotebook *notebook,
-                       gint abs_x, gint abs_y)
-{
-  EphyNotebook *nb_at_pointer;
-
-  nb_at_pointer = find_notebook_at_pointer (gtk_widget_get_display (GTK_WIDGET (notebook)),
-                                            abs_x, abs_y);
-
-  return nb_at_pointer == notebook;
-}
-
 static gint
 find_tab_num_at_pos (EphyNotebook *notebook, gint abs_x, gint abs_y)
 {
@@ -207,12 +180,6 @@ find_tab_num_at_pos (EphyNotebook *notebook, gint abs_x, gint abs_y)
   GtkNotebook *nb = GTK_NOTEBOOK (notebook);
   GtkWidget *page;
 
-  /* For some reason unfullscreen + quick click can
-     cause a wrong click event to be reported to the tab */
-  if (!is_in_notebook_window (notebook, abs_x, abs_y)) {
-    return NOT_IN_APP_WINDOWS;
-  }
-
   while ((page = gtk_notebook_get_nth_page (nb, page_num))) {
     GtkWidget *tab;
     GtkAllocation allocation;
@@ -310,7 +277,7 @@ notebook_drag_data_received_cb (GtkWidget        *widget,
   if (target == gdk_atom_intern_static_string ("GTK_NOTEBOOK_TAB"))
     return;
 
-  g_signal_stop_emission_by_name (widget, "drag_data_received");
+  g_signal_stop_emission_by_name (widget, "drag-data-received");
 
   if (g_settings_get_boolean (EPHY_SETTINGS_LOCKDOWN,
                               EPHY_PREFS_LOCKDOWN_ARBITRARY_URL)) return;
@@ -376,7 +343,8 @@ update_tabs_visibility (EphyNotebook *nb,
   mode = ephy_embed_shell_get_mode (EPHY_EMBED_SHELL (ephy_shell_get_default ()));
   num = gtk_notebook_get_n_pages (GTK_NOTEBOOK (nb));
 
-  if (before_inserting) num++;
+  if (before_inserting)
+    num++;
 
   policy = g_settings_get_enum (EPHY_SETTINGS_UI,
                                 EPHY_PREFS_UI_TABS_BAR_VISIBILITY_POLICY);
@@ -491,6 +459,41 @@ ephy_notebook_init (EphyNotebook *notebook)
                     G_CALLBACK (show_tabs_changed_cb), notebook);
 }
 
+static void
+ephy_notebook_constructed (GObject *object)
+{
+  EphyNotebook *notebook = EPHY_NOTEBOOK (object);
+  GtkWidget *hbox;
+  GtkWidget *button;
+  GtkWidget *popover;
+
+  G_OBJECT_CLASS (ephy_notebook_parent_class)->constructed (object);
+
+  hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
+  gtk_notebook_set_action_widget (GTK_NOTEBOOK (notebook), hbox, GTK_PACK_END);
+  gtk_widget_show (hbox);
+
+  button = gtk_button_new_from_icon_name ("tab-new-symbolic", GTK_ICON_SIZE_BUTTON);
+  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
+  gtk_actionable_set_action_name (GTK_ACTIONABLE (button), "win.new-tab");
+  /* Translators: tooltip for the new tab button */
+  gtk_widget_set_tooltip_text (button, _("Open a new tab"));
+  gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
+  gtk_widget_show (button);
+
+  button = gtk_menu_button_new ();
+  gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
+  /* Translators: tooltip for the tab switcher menu button */
+  gtk_widget_set_tooltip_text (button, _("View open tabs"));
+  gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
+  gtk_widget_show (button);
+
+  notebook->tab_menu = g_menu_new ();
+  popover = gtk_popover_new (button);
+  gtk_popover_bind_model (GTK_POPOVER (popover), G_MENU_MODEL (notebook->tab_menu), "win");
+  gtk_menu_button_set_popover (GTK_MENU_BUTTON (button), popover);
+}
+
 static void
 ephy_notebook_finalize (GObject *object)
 {
@@ -511,6 +514,86 @@ ephy_notebook_finalize (GObject *object)
   G_OBJECT_CLASS (ephy_notebook_parent_class)->finalize (object);
 }
 
+static const char *
+get_nth_tab_label_text (GtkNotebook *notebook,
+                        int          n)
+{
+  GtkWidget *page;
+  GtkWidget *tab_label;
+  GtkWidget *label;
+
+  g_assert (n >= 0);
+
+  page = gtk_notebook_get_nth_page (notebook, n);
+  g_assert (page != NULL);
+
+  tab_label = gtk_notebook_get_tab_label (notebook, page);
+  g_assert (GTK_IS_BOX (tab_label));
+
+  label = g_object_get_data (G_OBJECT (tab_label), "label");
+  g_assert (GTK_IS_LABEL (label));
+
+  return gtk_label_get_text (GTK_LABEL (label));
+}
+
+static char *
+ellipsize_tab_label (const char *label)
+{
+  static const int MAX_LENGTH = 50;
+  char *substring;
+  char *result;
+
+  if (g_utf8_strlen (label, -1) < MAX_LENGTH)
+    return g_strdup (label);
+
+  substring = g_utf8_substring (label, 0, MAX_LENGTH);
+  result = g_strconcat (substring, "…", NULL);
+  g_free (substring);
+
+  return result;
+}
+
+static void
+ephy_notebook_rebuild_tab_menu (EphyNotebook *notebook)
+{
+  GMenuItem *item;
+  const char *text;
+  char *ellipsized_text;
+  int num_pages;
+  GtkWidget *window;
+  GActionGroup *group;
+  GAction *action;
+  gint current_page;
+
+  g_menu_remove_all (notebook->tab_menu);
+
+  num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
+
+  /* TODO: Add favicon as well. Will have to ditch GMenu. :( */
+  for (int i = 0; i < num_pages; i++) {
+    text = get_nth_tab_label_text (GTK_NOTEBOOK (notebook), i);
+    ellipsized_text = ellipsize_tab_label (text);
+    item = g_menu_item_new (ellipsized_text, NULL);
+    g_menu_item_set_action_and_target (item, "show-tab", "u", (guint)i, NULL);
+    g_menu_append_item (notebook->tab_menu, item);
+    g_free (ellipsized_text);
+    g_object_unref (item);
+  }
+
+  current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));
+  if (current_page < 0)
+    return;
+
+  window = gtk_widget_get_toplevel (GTK_WIDGET (notebook));
+  group = gtk_widget_get_action_group (window, "win");
+  /* Is window being destroyed? */
+  if (group == NULL)
+    return;
+
+  action = g_action_map_lookup_action (G_ACTION_MAP (group), "show-tab");
+  g_simple_action_set_state (G_SIMPLE_ACTION (action), g_variant_new_uint32 ((guint32)current_page));
+}
+
 static void
 sync_load_status (EphyWebView *view, GParamSpec *pspec, GtkWidget *proxy)
 {
@@ -557,6 +640,14 @@ sync_label (EphyEmbed *embed, GParamSpec *pspec, GtkWidget *label)
   gtk_widget_set_tooltip_text (label, title);
 }
 
+static void
+rebuild_tab_menu_cb (EphyEmbed    *embed,
+                     GParamSpec   *pspec,
+                     EphyNotebook *notebook)
+{
+  ephy_notebook_rebuild_tab_menu (notebook);
+}
+
 static void
 sync_is_playing_audio (WebKitWebView *view,
                        GParamSpec    *pspec,
@@ -597,11 +688,9 @@ tab_label_style_set_cb (GtkWidget *hbox,
   char_width = pango_font_metrics_get_approximate_digit_width (metrics);
   pango_font_metrics_unref (metrics);
 
-  gtk_icon_size_lookup_for_settings (gtk_widget_get_settings (hbox),
-                                     GTK_ICON_SIZE_MENU, &w, &h);
+  gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &w, &h);
 
-  gtk_widget_set_size_request
-    (hbox, TAB_WIDTH_N_CHARS * PANGO_PIXELS (char_width) + 2 * w, -1);
+  gtk_widget_set_size_request (hbox, TAB_WIDTH_N_CHARS * PANGO_PIXELS (char_width) + 2 * w, -1);
 
   button = g_object_get_data (G_OBJECT (hbox), "close-button");
   gtk_widget_set_size_request (button, w + 2, h + 2);
@@ -637,7 +726,6 @@ build_tab_label (EphyNotebook *nb, EphyEmbed *embed)
   label = gtk_label_new (NULL);
   gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
   gtk_label_set_single_line_mode (GTK_LABEL (label), TRUE);
-  gtk_misc_set_padding (GTK_MISC (label), 0, 0);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
   gtk_widget_show (label);
 
@@ -651,7 +739,7 @@ build_tab_label (EphyNotebook *nb, EphyEmbed *embed)
   gtk_button_set_relief (GTK_BUTTON (close_button),
                          GTK_RELIEF_NONE);
   /* don't allow focus on the close button */
-  gtk_button_set_focus_on_click (GTK_BUTTON (close_button), FALSE);
+  gtk_widget_set_focus_on_click (close_button, FALSE);
 
   gtk_widget_set_name (close_button, "ephy-tab-close-button");
 
@@ -696,6 +784,8 @@ build_tab_label (EphyNotebook *nb, EphyEmbed *embed)
                            G_CALLBACK (sync_icon), icon, 0);
   g_signal_connect_object (embed, "notify::title",
                            G_CALLBACK (sync_label), label, 0);
+  g_signal_connect_object (embed, "notify::title",
+                           G_CALLBACK (rebuild_tab_menu_cb), nb, 0);
   g_signal_connect_object (view, "load-changed",
                            G_CALLBACK (load_changed_cb), box, 0);
   g_signal_connect_object (view, "notify::is-playing-audio",
@@ -843,6 +933,8 @@ ephy_notebook_remove (GtkContainer *container,
     (view, G_CALLBACK (sync_icon), tab_label_icon);
   g_signal_handlers_disconnect_by_func
     (tab_widget, G_CALLBACK (sync_label), tab_label_label);
+  g_signal_handlers_disconnect_by_func
+    (tab_widget, G_CALLBACK (sync_label), notebook);
   g_signal_handlers_disconnect_by_func
     (view, G_CALLBACK (sync_load_status), tab_label);
   g_signal_handlers_disconnect_by_func
@@ -853,6 +945,38 @@ ephy_notebook_remove (GtkContainer *container,
   update_tabs_visibility (notebook, FALSE);
 }
 
+static void
+ephy_notebook_page_added (GtkNotebook *notebook,
+                          GtkWidget   *child,
+                          guint        page_num)
+{
+  if (GTK_NOTEBOOK_CLASS (ephy_notebook_parent_class)->page_added != NULL)
+    GTK_NOTEBOOK_CLASS (ephy_notebook_parent_class)->page_added (notebook, child, page_num);
+
+  ephy_notebook_rebuild_tab_menu (EPHY_NOTEBOOK (notebook));
+}
+
+static void
+ephy_notebook_page_removed (GtkNotebook *notebook,
+                            GtkWidget   *child,
+                            guint        page_num)
+{
+  if (GTK_NOTEBOOK_CLASS (ephy_notebook_parent_class)->page_removed != NULL)
+    GTK_NOTEBOOK_CLASS (ephy_notebook_parent_class)->page_removed (notebook, child, page_num);
+
+  ephy_notebook_rebuild_tab_menu (EPHY_NOTEBOOK (notebook));
+}
+
+static void ephy_notebook_page_reordered (GtkNotebook *notebook,
+                                          GtkWidget   *child,
+                                          guint        page_num)
+{
+  if (GTK_NOTEBOOK_CLASS (ephy_notebook_parent_class)->page_reordered != NULL)
+    GTK_NOTEBOOK_CLASS (ephy_notebook_parent_class)->page_reordered (notebook, child, page_num);
+
+  ephy_notebook_rebuild_tab_menu (EPHY_NOTEBOOK (notebook));
+}
+
 /**
  * ephy_notebook_next_page:
  * @notebook: an #EphyNotebook
